"""A higher level API to provide a single function to the external merge sort
"""
from collections.abc import Iterator
from merge_sort import chunks, merge
import os
# import pprint as pp

_HEADER_KWARG = 'header'
"""The keyword argument name for the header value (`str`)
"""
_DELIMITER_KWARG = 'delimiter'
"""The keyword argument name for the csv delimiter value (`str`)
"""
_LINETERM_KWARG = 'lineterminator'
"""The keyword argument name for the csv lineterminator value (`str`)
"""
_CSV_KWARGS = 'csv_kwargs'
"""The keyword argument name for the csv keyword arguments (`str`)
"""


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def sorted(initer, key=None, reverse=False, chunksize=100000, tmpdir=None):
    """Perform an external merge sort on an iterator.

    Parameters
    ----------
    initer : `iterator`
        An iterator that provides, strings, lists or dicts to use in the sort.
        If the object is not an iterator an attempt will be made to make it
        into an iterator. An iterator is required as the first row is used to
        determine the chunker used for the sort. For list like objects will a
        csv based chunker will be used, for dicts a csv dict based chunker will
        be used. The keys of the dictionary will be used as a header row for the
        file, so every key will be converted to a string.
    key : `function`, optional, default: `NoneType`
        A function to provide the sort key. Please keep in mind that this will
        be used for both chunking the input and merging the intermediate files.
        When merging the data types of any list or dictionary values will be
        strings, so your function should perform appropriate casts for data
        types even if the input rows already have those data types.
    reverse : `bool`, optional, default: `False`
        Sort in reverse.
    chunksize : `int`, optional, default: `100000`
        The number of rows to hold in memory at anyone time during the chunking
        to intermediate files.
    tmpdir : `str`, optional, default: `NoneType`
        The temporary directory to store the intermediate chunk files.

    Yields
    ------
    sorted_rows : `str` or `list` or `dict`
        The output rows will match the data type input rows. Although the
        values within a list or dictionary will all be strings even if the
        input had differently typed values. Also, keep in mind that unless the
        input was list or dict (i.e. using csv) then the yielded rows will have
        newlines.

    Raises
    ------
    TypeError
        If the first row of the input is not a `str`, `int`, `float`, `list`
        or `dict`.
    """
    # Check if the input is truly an iterator
    if not isinstance(initer, Iterator):
        try:
            initer = iter(initer)
        except TypeError as e:
            raise TypeError(
                "not an iterator and can't be made into one"
            ) from e

    csv_delimiter = "\t"
    first_row = next(initer)

    # Sort out the chunker/merger required for the sort process
    chunker = None
    merger = None
    chunk_kwargs = dict(key=key, chunksize=chunksize, reverse=reverse)
    merge_kwargs = dict(key=key, tmpdir=tmpdir, reverse=reverse)
    proc_func = _dummy
    if isinstance(first_row, (str, int, float)):
        chunker = chunks.ExtendedChunks
        merger = merge.IterativeHeapqMerge
        chunk_kwargs[_HEADER_KWARG] = None
        merge_kwargs[_HEADER_KWARG] = False
        proc_func = _newline
    elif isinstance(first_row, (list, tuple)):
        chunker = chunks.CsvExtendedChunks
        merger = merge.CsvIterativeHeapqMerge
        chunk_kwargs[_HEADER_KWARG] = None
        chunk_kwargs[_DELIMITER_KWARG] = csv_delimiter
        chunk_kwargs[_LINETERM_KWARG] = os.linesep
        merge_kwargs[_HEADER_KWARG] = False
        merge_kwargs[_CSV_KWARGS] = dict(delimiter=csv_delimiter,
                                         lineterminator=os.linesep)
    elif isinstance(first_row, dict):
        chunker = chunks.CsvDictExtendedChunks
        merger = merge.CsvDictIterativeHeapqMerge
        chunk_kwargs[_HEADER_KWARG] = [str(i) for i in first_row.keys()]
        chunk_kwargs[_DELIMITER_KWARG] = csv_delimiter
        chunk_kwargs[_LINETERM_KWARG] = os.linesep
        merge_kwargs[_HEADER_KWARG] = True
        merge_kwargs[_CSV_KWARGS] = dict(delimiter=csv_delimiter,
                                         lineterminator=os.linesep)
    else:
        raise TypeError(f"can't handle input type: {type(first_row)}")

    with chunker(tmpdir, **chunk_kwargs) as c:
        c.add_row(proc_func(first_row))
        for i in initer:
            c.add_row(proc_func(i))

    with merger(c.chunk_file_list, **merge_kwargs) as m:
        for i in m:
            yield i


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def _dummy(x):
    """A dummy passthrough function.

    Parameters
    ----------
    x : `Any`
        Value to passthrough.

    Returns
    -------
    x : `Any`
        The given value.
    """
    return x


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def _newline(x):
    """Convert a simple type to a string and add a newline if it does not
    already exist.

    Parameters
    ----------
    x : `str` or `int` or `float`
        The value to convert.

    Returns
    -------
    str_x : `str`
        The converted value.
    """
    x = str(x)
    if x.endswith("\n"):
        return x
    return x + "\n"



"""Functions to get some example data
"""
from pyaddons import utils
import random


_SEED = 1984
"""The random seed for data generation (`int`)
"""

_DTYPE_MAP = {'str': str, 'int': int, 'float': float}
"""Mappings between data type names and their actual types (`dict`)
"""


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def create_data(has_header=True, ncols=10, nrows=10, use_dict=False,
                comment=None, comment_lines=10):
    """Create a test data set to work on chunking.

    Parameters
    ----------
    has_header : `bool`, optional, default: `\t`
        Write a header line.
    ncols : `int`, optional, default: `10`
        The number of columns to write.
    nrows : `int`, optional, default: `10`
        The number of rows to write.
    use_dict : `bool`, optional, default: `False`
        Should a `list` of `dict` be returned instead of a `list` of `list`.
    comment : `str`, optional, default: `NoneType`
        Should 10 comment character lines be placed before the start of the
        data. If skiplines > 0 these will be located after them.
    comment_lines : `int`, optional, default: `10`
        If comment is not None, this is the number of comment lines that will
        be added to the data.

    Returns
    -------
    test_data : `list` of (`list` or `dict`)
        The format of the test data will vary depending on the parameters. If
        ``has_header`` is ``True``, then the first element of the list will
        always be a list of strings. If use_dict is False then the remainder of
        the elements will be lists of mixed types depending on the seed. If
        use_dict is True then the remainder of the elements will be dicts with
        the header names as keys and the data as values.
    col_data_types : `list` of `str`
        The data types of the columns, if ``use_dict`` is ``False`` then this
        will be a list of data type names represented as strings, i.e.
        ``'str``. If ``use_dict`` is ``True`` then this will be a dict of
        column names as keys and data type names as values.
    """
    test_data = []

    # Add any comment lines
    if comment is not None:
        for i in range(comment_lines):
            test_data.append([f"{comment} {i}"])

    # Now add a header if we have one
    header = None
    if has_header is True:
        header = [f"column{i}" for i in range(ncols)]
        test_data.append([f"column{i}" for i in range(ncols)])

    data_funcs = [(str, get_string), (float, get_float), (int, get_integer)]
    for i in range(nrows):
        row = []
        col_data_types = []
        for j in range(ncols):
            random.seed(_SEED * j * 2)
            df = data_funcs[random.randint(0, len(data_funcs)-1)]
            row.append(df[1](_SEED * i * j))
            col_data_types.append(df[0].__name__)

        if use_dict is False:
            test_data.append([r for r in row])
        else:
            if has_header is False:
                raise ValueError("need a header for dict data")
            test_data.append(dict([(h, r) for h, r in zip(header, row)]))

    if use_dict is True:
        col_data_types = \
            dict([(h, c) for h, c in zip(header, col_data_types)])

    return test_data, col_data_types


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def get_string(seed=_SEED):
    """Get a text string for a random integer value between 1-5000 (`str`)
    """
    random.seed(seed)
    return utils.get_text_number(random.randint(1, 5000))


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def get_float(seed=_SEED):
    """Get a random float value between 0-1 (`float`)
    """
    random.seed(seed)
    return random.random()


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def get_integer(seed=_SEED):
    """Get a random integer value between 1-5000 (`int`)
    """
    random.seed(seed)
    return random.randint(1, 5000)


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def set_row_type(row, dtypes):
    """Convert the data types for a row.

    Parameters
    ----------
    row : `list` or `dict`
        The row to convert.
    dtypes : `list` of `str`
        String representations for the data types in each column.

    Returns
    -------
    converted_row : `list` or `dict`
        The row converted to the dtypes.
    """
    if isinstance(row, list):
        return [_DTYPE_MAP[dt](i) for dt, i in zip(dtypes, row)]
    elif isinstance(row, dict):
        return dict([(k, _DTYPE_MAP[dt](v)) for dt, k, v in
                     zip(dtypes, row.keys(), row.values())])
    else:
        raise TypeError(f"can't handle: {type(row)}")



"""Common functions used across modules.
"""
import csv

_CSV_KWARGS = [
    'delimiter',
    'doublequote',
    'escapechar',
    'lineterminator',
    'quotechar',
    'quoting',
    'skipinitialspace',
    'strict',
    'dialect',
    'restval',
    'extrasaction'
]
"""The possible csv keyword arguments (`list of `str`)
"""


# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
class _CsvWrapper(object):
    """A wrapper around csv to provide a basic file-like interface.

    Parameters
    ----------
    file : `File`
        A file object to wrap in a csv reader/writer.
    mode : `str`
        The csv mode, either ``r`` (reader) or ``w`` (writer).
    **csv_kwargs
        Keyword arguments that will be passed through to csv. They should be
        appropriate for (dict) readers/writers.
    """

    READ_MODE = 'r'
    """The read mode indicator (`str`)
    """
    WRITE_MODE = 'w'
    """The write mode indicator (`str`)
    """

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def __init__(self, file, mode="r", **csv_kwargs):
        self._file = None
        self._csv = None
        self._csv_kwargs = csv_kwargs

        self.file = file

        if mode not in [self.WRITE_MODE, self.READ_MODE]:
            raise ValueError(
                f"mode must be read ({self.READ_MODE}) or write "
                f"({self.WRITE_MODE})"
            )
        self._mode = mode

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def __enter__(self):
        """Entry into the context manager.
        """
        return self

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def __exit__(self, *args):
        """Exit from the context manager.
        """
        return self

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def __iter__(self):
        """The iterator.
        """
        return self

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def __next__(self):
        """Get the next row from the csv reader.
        """
        try:
            return next(self._csv)
        except TypeError as e:
            if self._csv is not None:
                raise
            else:
                raise IOError("did you forget to open") from e

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def open(self):
        """Initialise the csv file.

        Returns
        -------
        self
            The wrapper object.
        """
        return self

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def close(self):
        """Close the csv wrapper file object.
        """
        self._file.close()

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @property
    def name(self):
        """Get the file name (`str`).
        """
        return self._file.name

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @property
    def csv_kwargs(self):
        """Get the csv keyword arguments (`dict`).
        """
        return self._csv_kwargs

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @property
    def mode(self):
        """Get the opening mode (`str`).
        """
        return self._mode

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @property
    def csv(self):
        """Get the csv reader/writer object (`csv.reader` or `csv.writer` or
        `csv.DictReader` or `csv.DictWriter` or `NoneType`).
        """
        return self._csv

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @csv.setter
    def csv(self, csv):
        """Set the csv reader/writer object, can only be set once (`csv.reader`
        or `csv.writer` or `csv.DictReader` or `csv.DictWriter` or `NoneType`).
        """
        if self._csv is not None:
            raise ValueError("the csv object can only be set once")
        self._csv = csv

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @property
    def file(self):
        """Get the file object being wrapped (`File`).
        """
        return self._file

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @file.setter
    def file(self, file):
        """Set the file object to be wrapped, can only be set once (`File`).
        """
        if self._file is not None:
            raise ValueError("the file object can only be set once")
        self._file = file

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def read(self):
        """Read the entire file (`list` of (`list` or `dict`)).
        """
        return [i for i in self._csv]

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def readline(self):
        """Read a single line from the file (`list` or `dict`).
        """
        return next(self._csv)

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def write(self, row):
        """Write a single row to the file (`list` or `dict`).
        """
        return self._csv.writerow(row)


# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
class CsvWrapper(_CsvWrapper):
    """A wrapper around a file to provide a basic file-like interface to a
    ``csv.reader``/``csv.writer``

    Parameters
    ----------
    file : `File`
        A file object to wrap in a csv reader/writer.
    mode : `str`
        The csv mode, either ``r`` (reader) or ``w`` (writer).
    **csv_kwargs
        Keyword arguments that will be passed through to csv. They should be
        appropriate for (dict) readers/writers.
    """
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def open(self):
        """Initialise the ``csv.reader`` or ``csv.writer``.

        Returns
        -------
        self
            The wrapper object.
        """
        if self.mode == self.READ_MODE:
            self.csv = csv.reader(self.file, **self.csv_kwargs)
        else:
            self.csv = csv.writer(self.file, **self.csv_kwargs)
        return self


# @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
class CsvDictWrapper(_CsvWrapper):
    """A wrapper around a file to provide a basic file-like interface to a
    ``csv.DictReader``/``csv.DictWriter``

    Parameters
    ----------
    file : `File`
        A file object to wrap in a csv reader/writer.
    mode : `str`
        The csv mode, either ``r`` (reader) or ``w`` (writer).
    fieldnames : `list` of `str`
        The field names that will be passed to a ``csv.DictWriter`` if mode is
        ``w``.
    **csv_kwargs
        Keyword arguments that will be passed through to csv. They should be
        appropriate for (dict) readers/writers.
    """
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def __init__(self, *args, fieldnames=None, **kwargs):
        super().__init__(*args, **kwargs)
        self._fieldnames = fieldnames

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def open(self):
        """Initialise the ``csv.DictReader`` or ``csv.DictWriter``.

        Returns
        -------
        self
            The wrapper object.
        """
        if self.mode == self.READ_MODE:
            self.csv = csv.DictReader(self.file, **self.csv_kwargs)
        else:
            self.csv = csv.DictWriter(self.file, self._fieldnames,
                                      **self.csv_kwargs)
        return self


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def extract_csv_kwargs(kwargs):
    """Extract the CSV keyword arguments from a dictionary.

    Parameters
    ----------
    kwargs : `dict`
        A dictionary of keyword arguments potentially containing some keyword
        arguments for the csv constructor.

    Returns
    -------
    csv_kwargs : `dict`
        Any csv keyword arguments that have been extracted. This will be empty
        if none are found.
    """
    csv_kwargs = dict()
    for i in _CSV_KWARGS:
        try:
            csv_kwargs[i] = kwargs.pop(i)
        except KeyError:
            pass
    return csv_kwargs
